-- Show the geographic distance between traceroute hops

-- NOTE
-- MaxMind's GeoLite City database does a great job with subscriber nodes. It
-- Doesn't do such a great job with router nodes. You should probably use
-- traceroute-distance.lua instead.

-- Great Circle calculation. Adapted from 
-- http://stackoverflow.com/questions/365826/calculate-distance-between-2-gps-coordinates/365857#365857
local EARTH_RAD = 6378137.0 
-- earth's radius in meters (official geoid datum, not 20,000km / pi)

local radmiles = EARTH_RAD*100.0/2.54/12.0/5280.0;
-- earth's radius in miles

local multipliers = {
	radians = 1, miles = radmiles, mi = radmiles, feet = radmiles * 5280,
	meters = EARTH_RAD, m = EARTH_RAD, km = EARTH_RAD / 1000, 
	degrees = 360 / (2 * math.pi), min = 60 * 360 / (2 * math.pi)
}

function gcdist(pt1_lat, pt1_lon, pt2_lat, pt2_lon, units) -- return distance in radians or given units
	--- this formula works best for points close together or antipodal
	--- rounding error strikes when distance is one-quarter Earth's circumference
	--- (ref: wikipedia Great-circle distance)
	local pt1_lat_rad = math.rad(pt1_lat)
	local pt1_lon_rad = math.rad(pt1_lon)
	local pt2_lat_rad = math.rad(pt2_lat)
	local pt2_lon_rad = math.rad(pt2_lon)

	local sdlat = math.sin((pt1_lat_rad - pt2_lat_rad) / 2.0);
	local sdlon = math.sin((pt1_lon_rad - pt2_lon_rad) / 2.0);
	local res = math.sqrt(sdlat * sdlat + math.cos(pt1_lat_rad) * math.cos(pt2_lat_rad) * sdlon * sdlon);
	res = res > 1 and 1 or res < -1 and -1 or res
	res = 2 * math.asin(res);
	if units then
		return res * assert(multipliers[units])
	else
		return res
	end
end

do
	ip_src_f = Field.new("ip.src")
	ip_dst_f = Field.new("ip.dst")
	ip_id_f = Field.new("ip.id")
	ip_ttl_f = Field.new("ip.ttl")
	ip_src_lat_f = Field.new("ip.geoip.src_lat")
	ip_src_lon_f = Field.new("ip.geoip.src_lon")
	ip_src_city_f = Field.new("ip.geoip.src_city")

	local function geo_distance_tap()
		-- Declare the window we will use
		local tw = TextWindow.new("Traceroute Distance")

		-- Traceroute probes (IP packets with low TTL values)
		-- {IP src, IP dst, IP ID} = {IP TTL}
		local probes = {}

		-- Probe responses (ICMP TTL exceeded messages)
	        -- {IP src, IP dst, IP TTL} = {IP hop, hop lon, hop lat, hop city}
		local responses = {}

		-- Look for probes (low TTL values)
		local probe_tap = Listener.new(nil, "ip.ttl < 40");

		function get_2addr_num_key(src, dst, id)
			return string.format("%s|%s|%05d", src, dst, id)
		end

		-- This will be called once for each matching packet
		-- Gather up the Src, IP ID, TTL
		function probe_tap.packet(pinfo,tvb)
			local ip_src = ip_src_f()
			local ip_dst = ip_dst_f()
			local ip_id = ip_id_f()
			local ip_ttl = ip_ttl_f()

			local src = tostring(ip_src.value)
			local dst = tostring(ip_dst.value)
			local id = tostring(ip_id.value)
			local pkey = get_2addr_num_key(src, dst, id)
			local ttl = ip_ttl.value

			-- Last seen TTL wins
			probes[pkey] = ttl
		end

		-- this function will be called whenever a reset is needed
		-- e.g. when reloading the capture file
		function probe_tap.reset()
			tw:clear()
			probes = {}
		end

		-- Look for ICMP TTL exceeded messages
		local response_tap = Listener.new(nil, "icmp.type == 11");

		-- Called once for each matching packet
		-- Gather up the Src, Dst, Hop, Lon Lat
		function response_tap.packet(pinfo,tvb)
			local src_addrs = {ip_src_f()}
			local dst_addrs = {ip_dst_f()}
			local ip_ids = {ip_id_f()}
			local ip_ttl = ip_ttl_f()
			local ip_src_lat = ip_src_lat_f()
			local ip_src_lon = ip_src_lon_f()
			local ip_src_city = ip_src_city_f()

			-- We need two IP headers.
			if (# src_addrs < 2 or # dst_addrs < 2 or # ip_ids < 2) then
				tw:append("Not enough:" .. # src_addrs .. "\n")
				do return end
			end
			
			local src = tostring(src_addrs[2].value)
			local dst = tostring(dst_addrs[2].value)
			local id = tostring(ip_ids[2].value)
			local hop = tostring(src_addrs[1].value)
			local hop_lat = ip_src_lat.value
			local hop_lon = ip_src_lon.value
			local hop_city = nil
			
			if (ip_src_city) then
				hop_city = ip_src_city.value
			end

			local pkey = get_2addr_num_key(src, dst, id)
			local ttl = probes[pkey]
			if (ttl ~= nil) then
				local rkey = get_2addr_num_key(src, dst, ttl)
				responses[rkey] = {
					Src = src,
					Dst = dst,
					Hop = hop,
					Lat = hop_lat,
					Lon = hop_lon,
					City = hop_city}
			end
		end

		-- this function will be called once every few seconds to update our window
		function response_tap.draw(t)
			local last_run = nil
			local last_city = nil
			local last_lat
			local last_lon

			local path = {}
			for rkey in pairs(responses) do
				table.insert(path, rkey)
			end
			table.sort(path)

			for num, rkey in pairs(path) do
				local hop_data = responses[rkey]
				local src = hop_data.Src
				local dst = hop_data.Dst
				local ttl = hop_data.TTL
				local hop = hop_data.Hop
				local hop_lat = hop_data.Lat
				local hop_lon = hop_data.Lon
				local hop_city = hop_data.City
				local this_run = src .. "|" .. dst -- "a.b.c.e|w.x.y.z"
			
				if (this_run ~= last_run) then
					if (last_run ~= nil) then
						tw:append("\n")
					end
					last_run = this_run
					last_hop = nil
					tw:append(string.format("Path from %s to %s\n", src, dst))
					tw:append(string.format("Hop                                    Lat/Lon           City          Distance\n"))
				end
			
				tw:append(string.format("%-29s", hop))

				if (hop_city == nil) then -- Lat & lon won't be valid.
					tw:append(string.format("%17s", "(unknown)"))
				else
					tw:append(string.format("%8.3f %8.3f    %-15s", hop_lat, hop_lon, hop_city))
					if (last_city ~= nil) then
						tw:append(string.format("    %7.2f mi", gcdist(hop_lat, hop_lon, last_lat, last_lon, "mi")))
					end
					last_city = hop_city
					last_lat = hop_lat
					last_lon = hop_lon
				end
				tw:append("\n")
			end
		end

		-- this function will be called whenever a reset is needed
		-- e.g. when reloading the capture file
		function response_tap.reset()
			tw:clear()
			responses = {}
		end

		function remove()
			-- this way we remove the listener than otherwise will remain running indifinitelly
			probe_tap:remove();
			response_tap:remove();
		end

		retap_packets()

		-- Tell the window to call the remove() function when closed
		tw:set_atclose(remove)

	end

	-- using this function we register our fuction
	-- to be called when the user selects the Tools->Test->Packets menu
	register_menu("Lua Scripts/Traced Route Geo Distance", geo_distance_tap, MENU_TOOLS_UNSORTED)
end
